# BCS - the next level of ACS

This package contains some (hopefully) convenient wrappers around official
ALMA ACS Python libraries.

We aim here to make the interface to ACS libraries, classes and functions
as Pythonic as possible, without actually modifying ACS.

If you are an ALMA ACS developer and think our modifications are interesting for
ALMA ACS to adopt, please get in contact.

# Installation

    pip install -i https://test.pypi.org/simple/ bcs

# Tests

There are no tests, but we will use `pytest`.

# Usage

## Interacting with the ACS Manager

**with ACS**
```python
from Acspy.Clients.SimpleClient import PySimpleClient
client = PySimpleClient()
client.availableComponents()
# --> a list of [ComponentInfos, ...]
client.findComponents()
# --> a list of [names, ...]
```

**In BCS** we cut out the middleman and we introduce pythonic interfaces:
```python
from bcs import manager
manager.component_infos()
# a dict of {name: ComponentInfo, ...}
```

For more hopefully useful functions please have a look at `manager/impl.py` or simply tab-complete:
![bcs_manager_tab_complete.gif](docs/bcs_manager_tab_complete.gif)

### What if ACS is down?

BCS interfaces never block, when the ACS manager is unavailable. Instead they raise an exception.
It is up to the application developer to retry at a later time.

## Component development

<span style="color:red">
**Note:** The functionality below is not yet implemented.
</span>

BCS strives to make Python Component development super easy. So there is no need to implment XML Component simulators anymore.
Consider this IDL:

```C++
module example
{
    interface Amazing : ACS::ACSComponent {
        string speak (in double num);
        readonly attribute string last_word_spoken;
    };
};
```

The implementation of this component in Python will then be:
```python
# exampleImpl/Amazing.py
from bcs import Component

class Amazing(Component):
    last_word_spoken = ""

    def speak(self, num):
        self.last_word_spoken = str(num)
        return self.last_word_spoken
```

### Testability

In order to test the component implemented above, we might want two sets of tests:
 - tests, which do not need ACS to be up and running.
 - tests, with ACS.

Let's first consider tests without ACS being up:
```python
from exampleImpl import Amazing

def test_amazing():
    ama = Amazing()
    assert ama.speak(5) == "5"
    assert ama.last_word_spoken == "5"
```

For tests, like these to work, `Amazing()` must be constructable without ACS being up. This is not the case for normal ACS Components.

---

Let's now consider tests with ACS utilized. For this we need to consider a `CDB/MACI/Components/Components.xml` entry like this:
```xml
<e Name="AMA"
   Type="IDL:example/Amazing:1.0"
   Code="exampleImpl.Amazing"
   Container="some_container"
   ImplLang="py"
/>
```

We assume ACS is currently running and the Python `Container` named "some_container" has already been started.
Then we can execute tests like this:

```python
import bcs

def test_amazing():
    ama = bcs.manager.get_component(name="AMA")
    assert ama.speak(5) == "5"
    assert ama.last_word_spoken == "5"
```

## Logging

BCS embraces the design of the python std-library [`logging`](https://docs.python.org/3/library/logging.html) and does not try to break with it.
BCS encourages *library* and *component* developers to simply use the well known `logging` library.
It is up the the *application developer* to hook an `ACSLoggingHandler` into the application, if they wish to send log messages to the ACS loggin service.

Let's consider the case of a component developer, using the `Amazing` component again.
For them, logging should be a simple as this:

```python
# exampleImpl/Amazing.py
from bcs import Component
import logging

log = logging.getLogger(__name__)

class Amazing(Component):
    last_word_spoken = ""

    def speak(self, num):
        log.info("speak(num=%d)", num)
        self.last_word_spoken = str(num)
        return self.last_word_spoken
```

We see:
 - the developer followed the [good practice](https://docs.python.org/3/howto/logging.html#advanced-logging-tutorial)
   of creating a `logger` with the `__name__` of the module.
 - They used the std logging call `info()` instead of an ACS special call `logInfo()`.
 - They did not format the string inside the `info()` call, but allowed the logging library to defer string formatting until really necessary.

### Testabilty

If creating log messages is considered a feature of a function or class, then this feature should be tested!
Often logging remains untested, due to various reasons. So how do we test this in python?

We are making use of the [pytest fixture `caplog`](https://docs.pytest.org/en/stable/logging.html#caplog-fixture). For example like this:
```python
from exampleImpl import Amazing

def test_amazing(caplog):
    ama = Amazing()
    assert ama.speak(5) == "5"
    assert ama.last_word_spoken == "5"

    # below we assert that speak really emitted an INFO log message
    for record in caplog.records:
        assert record.levelname == "INFO"
    assert "speak(num" in caplog.text
```

How des BCS help us here? Well, the original `Acspy.Common.Log.Logger` class prevents the pytest fixture `caplog` from working properly. So when being forced to use the ACS Logger class, properly testing logging was made impossible.
