"""
Convenience Wrapper for ALMA ACS
Copyright (C) 2021  Dominik Neise

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import atexit
from functools import wraps

from Acspy.Clients.SimpleClient import PySimpleClient
from Acspy.Util.ACSCorba import getManager


__client = None


def invalidate_client_upon_exception(f):
    @wraps(f)
    def wrapper(*args, **kwds):
        try:
            return f(*args, **kwds)
        except:
            global __client
            if __client is not None:
                # NOTE TODO:
                # invalidating the client for all exceptions is not correct.
                # some exceptions are not raised due to the manager going offline
                # but due to programmer errors, so we invalidate the client, while the
                # manager is still up and the connection is totally fine.
                atexit.unregister(__client.disconnect)
            __client = None
            raise

    return wrapper


def simple_client(block=False):
    """get PySimpleClient without blocking"""
    if getManager() is None and not block:
        raise ConnectionError("Manager unavailable - client cannot log in")

    return PySimpleClient("acada_acs__client")


def is_available():
    """return True if ACS Manager is available"""
    return getManager() is not None


def __get_client():
    global __client
    if __client is None:
        __client = simple_client(block=False)
    return __client


# NOTE:
# The function below is not is use yet, but the 3 lines in the body
# are one way to "re-activate" a client after ACS went down and up again
# (took me a while to figure it out .. so I'd like to keep it here for a moment)
# def __reconnect_client():
#     c = __client
#     c.managerLogout()  # just to be nice
#     c.token = c.managerLogin()
#     c.setAll(c.name, c.token, c.token.h, c._PySimpleClient__activateOffShoot)
#


@invalidate_client_upon_exception
def component_infos(name_matches="*", type_matches="*", only_activated=False):
    """dict {name: ComponentInfo} consisting of all Components known
    to manager.

    Parameters:
    - name_matches is a wildcard that the components name must match in
    order for its information to be returned
    - type_matches is a wildcard that the components type must match in
    order for its information to be returned.
    - activated is a boolean value which specifies whether the search
    should be limited to components which have already been activated.

    Returns: a list consisting of ComponentInfo structures for every
    component manager knows of

    Raises: if manager is not available
    """
    c = __get_client()

    _component_infos = getManager().get_component_info(
        c.token.h, [], name_matches, type_matches, only_activated
    )
    return {ci.name: ci for ci in _component_infos}

# TODO: Container and Client Infos are currently not available
# since our PySimpleClient does not have the right permissions
#
# @invalidate_client_upon_exception
# def container_infos():
#     """dict {name: ContainerInfo} consisting of all ContainerInfos known
#     to manager.
#
#     Raises: if manager is not available
#     """
#     c = __get_client()
#
#     _container_infos = getManager().get_container_info(c.token.h, [], "*")
#     return {ci.name: ci for ci in _container_infos}
#
# @invalidate_client_upon_exception
# def client_infos():
#     """dict {name: ClientInfo} consisting of all ClientInfos known
#     to manager.
#
#     Raises: if manager is not available
#     """
#     c = __get_client()
#
#     _client_infos = getManager().get_client_info(c.token.h, [], "*")
#     return {ci.name: ci for ci in _client_infos}


@invalidate_client_upon_exception
def restart_component(name):
    """Restart a component.

    name(str): name of the component

    Returns: Reference to the restarted component.

    Raises: if manager is not available
    """
    c = __get_client()
    return getManager().restart_component(c.token.h, name)


@invalidate_client_upon_exception
def set_mortality(name, immortal=True):
    """change mortality state of a component.
    Component must be already active.
    The caller must be owner of the component or have administator rights.

    Parameters:
    name(str): name of the component
    immortal(bool): if True - make component immortal
                    if False - make component mortal again.

    Raises: if manager is not available
    """
    c = __get_client()
    getManager().make_component_immortal(c.token.h, name, immortal)


@invalidate_client_upon_exception
def release_component(name, force=False):
    """
    Release the component defined by name.

    Parameter:
    name(str): name of the component to be released.
    force(bool): releases forcefully if set to True

    Returns: The number of objects still attached to the released component

    Raises: if manager is not availabe
    """
    c = __get_client()
    if force:
        return getManager().force_release_component(c.token.h, name)

    return getManager().release_component(c.token.h, name)


@invalidate_client_upon_exception
def get_component(
    name=None,
    activate=True,
    idl_type=None,
    comp_code=None,
    container_name=None,
    is_dynamic=False,
    immortal=False,
):
    c = __get_client()
    comp = c.getComponent(
        name, activate, idl_type, comp_code, container_name, int(is_dynamic)
    )
    if immortal:
        set_mortality(comp.name, True)

    return comp


def try_activate_component(
    name=None,
    activate=True,
    idl_type=None,
    comp_code=None,
    container_name=None,
    is_dynamic=False,
    immortal=False,
):
    try:
        return get_component(
            name, activate, idl_type, comp_code, container_name, is_dynamic, immortal
        )
    except:
        # TODO: what should we do? log, warn, print?
        return None


def try_release_immortal_by_name(name):
    #TODO:
    # if `name` contains a typo, we get the exact same exception as if
    # we asked to release a compoent which was already relesaed earlier or never activated
    # so typos result in the same behavior as valid use-cases ... NOT GOOD

    # TODO: should we at least return a boolean telling the user if it worked?

    try:
        get_component(name, activate=False)
    except:
        # component is not active, no need to release
        return

    set_mortality(name, False)
    release_component(name)


def try_release_immortal_by_prefix(prefix):
    try:
        for comp_name in component_infos(name_matches=f"{prefix}*").keys():
            try_release_immortal_by_name(comp_name)
    except:
        return
